// Copyright 1999-2005, 2011 Jon Dart. All Rights Reserved.

#include "hash.h"
#include "util.h"
#include "debug.h"
#include "constant.h"
#include "globals.h"
#include "legal.h"
#include "learn.h"
#include "scoring.h"
extern "C"
{
#ifndef _MAC
#include <malloc.h>
#endif
#include <memory.h>
#include <stddef.h>
};

static DualHashEntry *hashTable = NULL;
int hashSize;
static const int MaxRehash = 3;
static int refCount = 0;
static int hash_init_done = 0;

void initHash(size_t bytes)
{
   int hash_entries;
   hash_entries = (int)(3*(bytes/sizeof(HashEntry))/4);
   if (!hash_init_done) {
                                                  // ensure it's even
      hashSize = hash_entries/2 + (hash_entries/2 % 2);
      ALIGNED_MALLOC(hashTable,
         DualHashEntry,
         sizeof(DualHashEntry)*hashSize,128);
      Scoring::initEvalCache(bytes/4);
      clearHash();
      hash_init_done++;
      ++refCount;
   }
}


void resizeHash(size_t bytes)
{
   freeHash();
   initHash(bytes);
}


void freeHash()
{
   if (--refCount == 0) {
      ALIGNED_FREE(hashTable);
      Scoring::freeEvalCache();
      hash_init_done = 0;
   }
}


void clearHash()
{
   memset(hashTable,'\0',sizeof(DualHashEntry)*hashSize);
   Scoring::clearEvalCache();
}


void loadLearnInfo()
{
   if (options.learning.position_learning) {
      ifstream plog;
      plog.open(learnFileName.c_str(),ios_base::in);
      while (plog.good() && !plog.eof()) {
         LearnRecord rec;
         if (getLearnRecord(plog,rec)) {
            Move best = NullMove;
            if (rec.start != InvalidSquare)
               best = CreateMove(rec.start,rec.dest,rec.promotion);
            HashEntry pos(rec.hashcode,rec.depth*DEPTH_INCREMENT,
               0,                                 /* age */
               PositionInfo::Valid,
               rec.in_check,
               rec.score,1,
               best);
            storeHash(pos, rec.hashcode);
         }
      }
   }
}


PositionInfo::ValueType searchHash(const Board& b,hash_t hashCode,
int alpha, int beta, int ply,
int depth, int age,
PositionInfo &pi
)
{
   int i;
   int probe = quickmod(hashCode,hashSize);
   DualHashEntry *p = &hashTable[probe];
   HashEntry *hit = NULL;
   if (p->replace.getEffectiveHash() == hashCode) {
      pi = p->replace;
      hit = &p->replace;
      if (p->replace.depth() >= depth) {
         // usable depth
         goto hash_hit;
      }
      // got a hit but depth not sufficient: keep the
      // data, but try the depth table too
   }
   for (i = MaxRehash; i != 0; --i) {
      HashEntry &entry = p->depth;
      if (entry.getEffectiveHash() == hashCode) {
        /*
         // we got a hit on this entry in the current search,
         // so update the age to discourage replacement:
         if (entry.age() && (entry.age() != age)) {
            entry.setAge(age);
            entry.setEffectiveHash(hashCode);
         }
        */
         pi = entry;
         hit = &entry;
         if (entry.depth() >= depth) {
            // usable depth
            goto hash_hit;
         }
         else {
            break;
         }
      }
      probe++; p++;
      if (probe >= hashSize) {
         probe = 0; p = hashTable;
      }
   }
   // If we got a hash hit but insufficient depth return here:
   if (hit) return PositionInfo::Invalid;
   return PositionInfo::NoHit;
   hash_hit:
   //
   // Whether we can use the hash value or not depends on
   // the flags:
   //
   int value = pi.value();
   switch (pi.type()) {
      case PositionInfo::Valid:
         // If this is a mate score, adjust it to reflect the
         // current ply depth.
         //
         if (value >= Constants::MATE_RANGE) {
            value -= ply;
            pi.setValue(value);
         }
         else if (value <= -Constants::MATE_RANGE) {
            value += ply;
            pi.setValue(value);
         }
         return PositionInfo::Valid;
      case PositionInfo::LowerBound:
         if (value >= beta)
            return PositionInfo::LowerBound;
         break;
      case PositionInfo::UpperBound:
         if (value <= alpha)
            return PositionInfo::UpperBound;
         break;
      default:
         break;
   }                                              // end switch
   return PositionInfo::Invalid;
}


void storeHash(const HashEntry &newPos, hash_t hashCode,
Statistics *stats)
{
   int probe = quickmod(hashCode,hashSize);
   DualHashEntry *p = &hashTable[probe];
   p->replace = newPos;

   HashEntry *best = NULL;
   ASSERT(Util::Abs(newPos.value()) <= Constants::MATE);
   // Of the positions that hash to the same locations
   // as this one, find the best one to replace.
   int minDepth = 10000;
   for (int i = MaxRehash; i != 0; --i) {
      HashEntry &q = p->depth;

      if (q.hashCode() == (hash_t)0 || (q.age() &&
                                        q.age() != newPos.age())) {
         best = &q;
         break;
      }
      else if (q.getEffectiveHash() == hashCode) {
        // same position
        if (q.depth() <= newPos.depth()) {
            best = &q;
            break;
         }
      }
      else if (q.depth() < minDepth && q.age()) {
        minDepth = q.depth();
        if (newPos.depth() >= minDepth) best = &q;
      }

      probe++; p++;
      if (probe >= hashSize) {
         probe = 0; p = hashTable;
      }
   }
   if (best != NULL) {
      if (stats) {
         if (best->hashCode() == (hash_t)0L)
            stats->hash_inserts++;
         else
            stats->hash_replaces++;
      }
      *best = newPos;
   }
   else if (stats) {
      stats->hash_inserts_failed++;
   }
}
